#!/havocbot

from havocbot.plugin import HavocBotPlugin
import logging

logger = logging.getLogger(__name__)


class HelpPlugin(HavocBotPlugin):

    @property
    def plugin_description(self):
        return "basics of the bot"

    @property
    def plugin_short_name(self):
        return "help"

    @property
    def plugin_usages(self):
        return [
            ("!help", None, "display basic help information"),
            ("!help plugins", None, "list the enabled plugins"),
            ("!help plugin <plugin>", None, "display detailed information about a specific plugin"),
        ]

    @property
    def plugin_triggers(self):
        return [
            ("!help", self.start),
            ("!help plugins", self.help_plugins),
            ("!help plugin", self.help_plugin),
        ]

    def init(self, havocbot):
        self.havocbot = havocbot

        # This will register the above triggers with havocbot
        self.havocbot.register_triggers(self.plugin_triggers)

    def shutdown(self):
        self.havocbot.unregister_triggers(self.plugin_triggers)
        self.havocbot = None

    def start(self, callback, message, **kwargs):
        callback_message_list = ["HavocBot can help you with the following."]
        for usage_message in self.get_usage_lines_from_plugin_as_list(self.plugin_usages):
            callback_message_list.append(usage_message)

        if message.channel:
            callback.send_messages_from_list(channel=message.channel, message=callback_message_list)

    def help_plugins(self, callback, message, **kwargs):
        callback_message_list = []

        tuples_list = self.get_help_list_sorted()

        callback_message_list.append("Loaded Plugins - " + ", ".join([("%s (%s)" % (i[1], i[0])) for i in tuples_list]))
        callback_message_list.append("Details about a specific plugin are available with '!help plugin <plugin>'")

        if message.channel:
            callback.send_messages_from_list(channel=message.channel, message=callback_message_list)

    def help_plugin(self, callback, message, **kwargs):
        callback_message_list = []

        tuples_list = self.get_help_list_sorted()
        words = message.text.split()

        if len(words) >= 3:
            temp_list = []
            for word in words[2:]:
                temp_list += [item for item in tuples_list if word in [item[0], item[1]]]
            if temp_list is not None and len(temp_list) > 0:
                callback_message_list = self.get_plugin_help_matching_list(temp_list)
            else:
                callback_message_list.append("No matching plugins were found. Plugin names are listed at !help plugins")
        else:
            # Print the plugin short name with the regular name in parenthesis as a joined string
            callback_message_list.append("Loaded Plugins - " + ", ".join([("%s (%s)" % (i[1], i[0])) for i in tuples_list]))
            callback_message_list.append("Details about a specific plugin are available with '!help plugin <plugin>'")

        if message.channel:
            callback.send_messages_from_list(channel=message.channel, message=callback_message_list)

    def get_plugin_help_matching_list(self, tuples_list):
        new_list = []

        for (plugin_name, plugin_short_name, dictionary) in tuples_list:
            if dictionary is not None:
                if 'description' in dictionary:
                    new_list.append("%s (%s) - %s" % (plugin_short_name, plugin_name, dictionary['description']))
                if 'usage' in dictionary:
                    for item in self.get_usage_lines_from_plugin_as_list(dictionary['usage']):
                        new_list.append(item)
            else:
                new_list.append("%s" % (plugin_short_name, plugin_name))

        return new_list

    def get_usage_lines_from_plugin_as_list(self, usages_tuples):
        usage_list = []

        for (usage, example, description) in usages_tuples:
            if usage is not None:
                if description is not None:
                    if example is not None:
                        message = "    Usage: %s - %s - (example '%s')" % (usage, description, example)
                        usage_list.append(message)
                    else:
                        message = "    Usage: %s - %s" % (usage, description)
                        usage_list.append(message)
                else:
                    message = "    Usage: %s" % (usage)
                    usage_list.append(message)

        return usage_list

    def get_help_list_sorted(self):
        tuples_list = []

        for statefulplugin in self.havocbot.plugins_core:
            help_tuple_item = self.get_help_item_tuple(statefulplugin.handler)
            tuples_list.append(help_tuple_item)

        for statefulplugin in self.havocbot.plugins_custom:
            help_tuple_item = self.get_help_item_tuple(statefulplugin.handler)
            tuples_list.append(help_tuple_item)

        # Sort in place with the first item in the tuple being the key
        tuples_list.sort(key=lambda help_item: help_item[1])

        return tuples_list

    # Returns a tuple that contains two items. First item is the name of the plugin as a string. Second item is a dictionary containing usage tuples, description, example, and the short name
    def get_help_item_tuple(self, plugin):
        help_tuple_item = ()

        if plugin is not None:
            # Capture the name of the plugin
            plugin_name = type(plugin).__name__
            plugin_short_name = plugin.plugin_short_name

            plugin_item_dictionary = {}

            if plugin.plugin_description is not None:
                plugin_item_dictionary['description'] = plugin.plugin_description

            # if plugin.plugin_short_name is not None:
            #     plugin_item_dictionary['short_name'] = plugin.plugin_short_name

            if plugin.plugin_usages is not None:
                plugin_item_dictionary['usage'] = plugin.plugin_usages

            help_tuple_item = (plugin_name, plugin_short_name, plugin_item_dictionary)

        return help_tuple_item


# Make this plugin available to HavocBot
havocbot_handler = HelpPlugin()



#!/havocbot

from havocbot.plugin import HavocBotPlugin
import logging

logger = logging.getLogger(__name__)


class RestartPlugin(HavocBotPlugin):

    @property
    def plugin_description(self):
        return "Restart the bot"

    @property
    def plugin_short_name(self):
        return "restart"

    @property
    def plugin_usages(self):
        return [
            ("!restart", None, "shut it down!"),
        ]

    @property
    def plugin_triggers(self):
        return [
            ("!restart", self.start),
        ]

    def init(self, havocbot):
        self.havocbot = havocbot

        # This will register the above triggers with havocbot
        self.havocbot.register_triggers(self.plugin_triggers)

    def shutdown(self):
        self.havocbot.unregister_triggers(self.plugin_triggers)
        self.havocbot = None

    def start(self, callback, message, **kwargs):
        if message.channel:
            callback.send_message(channel=message.channel, message="Restarting the bot. Hang tight")
        self.havocbot.restart()


# Make this plugin available to HavocBot
havocbot_handler = RestartPlugin()



from havocbot.client import Client
from havocbot.message import Message
from havocbot.user import User
import json
import logging
import re
from slackclient import SlackClient

logger = logging.getLogger(__name__)


class Slack(Client):

    @property
    def integration_name(self):
        return "slack"

    def __init__(self, havocbot):
        # Capture a reference to havocbot
        self.havocbot = havocbot

        self.token = None
        self.username = None
        self.user_id = None
        self.exact_match_one_word_triggers = True

    # Takes in a list of kv tuples in the format [('key', 'value'),...]
    def configure(self, settings):
        requirements_met = False

        for item in settings:
            # Switch on the key
            if item[0] == 'api_token':
                self.token = item[1]
                requirements_met = True
            elif item[0] == 'admins':
                pass

        # Return true if this integrations has the information required to connect
        if requirements_met:
            return True
        else:
            logger.error("Slack configuration is not valid. Check your settings and try again")
            return False

    def connect(self):
        if not self.token:
            logger.error("A Slack api token must be configured. Get one at https://slack.com/integrations")
            return False

        self.client = SlackClient(self.token)

        if self.client.rtm_connect():
            # Set some values from the login_data response
            self.username = self.client.server.login_data["self"]["name"]
            self.user_id = self.client.server.login_data["self"]["id"]

            logger.info("I am.. %s! (%s)" % (self.username, self.user_id))
            return True
        else:
            return False

    def disconnect(self):
        if self.client is not None:
            self.client.server.websocket.close()
            self.client = None

    def process(self):
        while not self.havocbot.should_shutdown:
            try:
                for event in self.client.rtm_read():
                    if 'type' in event:
                        if event['type'] == 'message':
                            logger.debug("raw message received - '%s'" % (event))

                            # Ignore messages originating from havocbot
                            if 'user' in event and event['user'] != self.user_id:
                                message_object = create_message_object_from_json(event)
                                logger.info("Received - %s" % (message_object))

                                try:
                                    self.handle_message(message_object=message_object)
                                except Exception as e:
                                    logger.error("Unable to handle the message")
                                    logger.error(e)
                        elif event['type'] == 'user_typing':
                            logger.debug("user_typing received - '%s'" % (event))
                        elif event['type'] == 'hello':
                            logger.debug("hello received - '%s'" % (event))
                        elif event['type'] == 'presence_change':
                            logger.debug("presence_change received - '%s'" % (event))
                        elif event['type'] == 'status_change':
                            logger.debug("status_change received - '%s'" % (event))
                        elif event['type'] == 'reconnect_url':
                            logger.debug("reconnect_url received - '%s'" % (event))
                        else:
                            logger.debug("UNKNOWN EVENT received - '%s'" % (event))
                    elif 'reply_to' in event:
                        logger.debug("reply_to received - '%s'" % (event))
                    else:
                        logger.debug("UNKNOWN THING received - '%s'" % (event))
            except AttributeError as e:
                logger.error("We have a problem! Is there a client?")
                logger.error(e)

    def handle_message(self, **kwargs):
        if kwargs is not None:
            if 'message_object' in kwargs and kwargs.get('message_object') is not None:
                message_object = kwargs.get('message_object')

            if message_object.type_ == 'message':
                for (trigger, triggered_function) in self.havocbot.triggers:
                    # Add exact regex match if user defined
                    if len(trigger.split()) == 1 and self.exact_match_one_word_triggers is True:
                        if not trigger.startswith('^') and not trigger.endswith('$'):
                            logger.debug("Converting trigger to a line exact match requirement")
                            trigger = "^" + trigger + "$"

                    # Use trigger as regex pattern and then search the message for a match
                    regex = re.compile(trigger)

                    match = regex.search(message_object.text)
                    if match is not None:
                        logger.info("Matched message against trigger '%s'" % (trigger))

                        # Pass the message to the function associated with the trigger
                        try:
                            triggered_function(self, message_object, capture_groups=match.groups())
                        except Exception as e:
                            logger.error(e)

                        break
                    else:
                        logger.debug("No match. Skipping '%s'" % (trigger))
                        pass
            else:
                logger.debug("Ignoring non message event of type '%s'" % (message_object.type_))

    def send_message(self, **kwargs):
        if kwargs is not None:
            if 'message' in kwargs and kwargs.get('message') is not None:
                message = kwargs.get('message')
            if 'channel' in kwargs and kwargs.get('channel') is not None:
                channel = kwargs.get('channel')

            if channel and message:
                logger.info("Sending message '%s' to channel '%s'" % (message, channel))
                try:
                    self.client.rtm_send_message(channel, message)
                except AttributeError:
                    logger.error("Unable to send message. Are you connected?")
                except Exception as e:
                    logger.error("Unable to send message. %s" % (e))

    def send_messages_from_list(self, **kwargs):
        if kwargs is not None:
            if 'message' in kwargs and kwargs.get('message') is not None:
                message = kwargs.get('message')
            if 'channel' in kwargs and kwargs.get('channel') is not None:
                channel = kwargs.get('channel')

            if channel and message:
                joined_message = "\n".join(message)
                logger.info("Sending message list '%s' to channel '%s'" % (joined_message, channel))
                try:
                    self.client.rtm_send_message(channel, joined_message)
                except AttributeError:
                    logger.error("Unable to send message. Are you connected?")
                except Exception as e:
                    logger.error("Unable to send message. %s" % (e))

    def get_user_by_id(self, user_id):
        user = None

        if self.client:
            user_json = self.client.api_call("users.info", user=user_id)
            data = json.loads(user_json.decode())

            user = create_user_object_from_json(data)
        else:
            logger.error("Unable to connect to the slack client")

        if user:
            return user
        else:
            return None


class SlackMessage(Message):
    def __init__(self, text, user, channel, type_, team, reply_to, ts):
        self.text = text
        self.user = user
        self.channel = channel
        self.type_ = type_
        self.team = team
        self.reply_to = reply_to
        self.ts = ts

    def __str__(self):
        return "SlackMessage(Text: '%s', User: '%s', Channel: '%s', Type: '%s', Team: '%s', Reply To: '%s', Timestamp: '%s')" % (self.text, self.user, self.channel, self.type_, self.team, self.reply_to, self.ts)


class SlackUser(User):
    def __init__(self, id, name, real_name, tz):
        self.id = id
        self.name = real_name
        self.username = name
        self.tz = tz

    def __str__(self):
        return "SlackUser(Name: '%s', Username: '%s', ID: '%s', Timezone: '%s')" % (self.name, self.username, self.id, self.tz)


def create_message_object_from_json(json):
    text = json['text'] if 'text' in json else None
    user = json['user'] if 'user' in json else None
    channel = json['channel'] if 'channel' in json else None
    type_ = json['type'] if 'type' in json else None
    team = json['team'] if 'team' in json else None
    reply_to = json['reply_to'] if 'reply_to' in json else None
    ts = json['ts'] if 'ts' in json else None

    message = SlackMessage(text, user, channel, type_, team, reply_to, ts)

    return message


# Returns a newly created user from a json source
def create_user_object_from_json(json):
    if 'real_name' in json['user'] and len(json['user']['real_name']) > 0:
        real_name = json['user']['real_name']
    else:
        real_name = None

    user = SlackUser(json['user']['id'], json['user']['name'], real_name, json['user']['tz'])

    return user



import logging

logger = logging.getLogger('__name__')


def catch_exceptions(function):
    def encapsulated_function(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as e:
            logger.exception(e)
            return False

    return encapsulated_function



#!/havocbot

from havocbot.plugin import HavocBotPlugin
import logging

logger = logging.getLogger(__name__)


class ReloadPlugin(HavocBotPlugin):

    @property
    def plugin_description(self):
        return "reloads all plugins"

    @property
    def plugin_short_name(self):
        return "reload"

    @property
    def plugin_usages(self):
        return [
            ("!reload", None, "shutdown and start all the plugins"),
        ]

    @property
    def plugin_triggers(self):
        return [
            ("!reload", self.start),
        ]

    def init(self, havocbot):
        self.havocbot = havocbot

        # This will register the above triggers with havocbot
        self.havocbot.register_triggers(self.plugin_triggers)

    def shutdown(self):
        self.havocbot.unregister_triggers(self.plugin_triggers)
        self.havocbot = None

    def start(self, callback, message, **kwargs):
        self.havocbot.reload_plugins()
        if message.channel:
            callback.send_message(channel=message.channel, message="Reloaded %d modules and discovered %s commands" % (len(self.havocbot.plugins_core) + len(self.havocbot.plugins_custom), len(self.havocbot.triggers)))


# Make this plugin available to HavocBot
havocbot_handler = ReloadPlugin()



import copy
from havocbot import pluginmanager
import inspect
import logging
import sys
import threading
import time

# Python2/3 compat
try:
    from queue import Queue
except ImportError:
    from Queue import Queue

logger = logging.getLogger(__name__)


class HavocBot:
    def __init__(self):
        self.clients = []
        self.queue = Queue()
        self.plugin_dirs = []
        self.plugins_core = []
        self.plugins_custom = []
        self.triggers = []
        self.settings = {}
        self.is_configured = False
        self.processing_threads = []
        self.should_shutdown = False
        self.should_restart = False

    def set_settings(self, **kwargs):
        self.settings['havocbot'] = kwargs.get('havocbot_settings', None)
        self.settings['clients'] = kwargs.get('clients_settings', None)

        self.configure()

    def configure(self):
        self.configure_bot(self.settings['havocbot'])
        self.configure_clients(self.settings['clients'])

        self.load_plugins()

        # The bot is now configured
        logger.debug("HavoceBot instance has been configured")
        self.is_configured = True

    def configure_bot(self, settings_dict):
        """ Configures the bot prior to starting up.

        Takes in a dictionary with keys relating to the bot name with the value
        containg a standard SafeConfigParser.items() tuple list

        example:
        {'havocbot': [('plugins_dir', 'plugins'), ('property2', 'value12)], ...}
        """
        if settings_dict is not None and 'havocbot' in settings_dict:
            for (key, value) in settings_dict['havocbot']:
                if key == 'plugin_dirs':
                    self.plugin_dirs = value.strip().split(",")

    def configure_clients(self, clients_dict):
        """ Configures a client integration prior to starting up.

        Takes in a dictionary with keys relating to the client name with the values
        containg a standard SafeConfigParser.items() tuple list. The client
        names are iterated over and a instance of the client integration is
        instantiated if possible. The tuple list is then passed to the configure()
        method inside the client integration for processing

        example:
        {'slask': [('plugins_dir', 'plugins'), ('property2', 'value12)], ...}
        """

        # Placeholder for client integration instances that will be override the self.clients value
        clientList = []

        if clients_dict is not None:
            # for client_name, client_settings_tuple_list in clients_dict.iteritems():  # Python 3 incompatible
            for client_name, client_settings_tuple_list in clients_dict.items():
                client_temp = self.import_and_return_client(client_name)
                if client_temp is not None:
                    # Instantiates the temp client and passes it the running HavocBot instance
                    client = client_temp(self)

                    if client.configure(client_settings_tuple_list):
                        clientList.append(client)

        # Override the client list with the new temp list
        self.clients = clientList

    # Takes in a name of a module and tries to import it
    def import_and_return_client(self, name, module=None):
        mod = self.import_module(name, module=None)

        # Inspect the module and iterate over all the members looking for entries that are classes that inherit from client.Client
        for name, obj in inspect.getmembers(mod):
            if inspect.isclass(obj):
                base_classes = inspect.getmro(obj)
                # Get the parent which should always appear as the second item
                parent = base_classes[1]
                if parent is not None:
                    if parent.__name__ == 'Client':
                        # logger.debug("Name: %s, Object: %s, Base Classes: %s" % (name, obj, base_classes))
                        client_instantiator = name

                        if client_instantiator is not None and hasattr(mod, client_instantiator):
                            client = getattr(mod, client_instantiator)
                            return client

        return None

    def import_module(self, name, module=None):
        try:
            if module is None:
                module = "havocbot.clients.%s" % (name)
                __import__(module)
                mod = sys.modules[module]

                return mod
        except ImportError:
            logger.error("Unable to import the %s client integration file" % (name))

            return None

    def load_plugins(self):
        self.plugins_core = pluginmanager.load_plugins_core(self)
        self.plugins_custom = pluginmanager.load_plugins_custom(self)

    def start(self):
        if self.is_configured is not True:
            sys.exit('Havocbot has not been configured. Please configure the bot and try again')
        else:
            logger.info("Starting HavocBot")

        if self.clients is not None and len(self.clients) > 0:
            logger.debug("Setting should_shutdown to False")
            self.should_shutdown = False

            # Connect and begin processing for each client in tuple
            for client in self.clients:
                # Spawn a thread for each unconnected client
                logger.debug("Spawning new daemon thread for client %s" % (client.integration_name))
                t = ClientThread(self)
                t.daemon = True
                self.processing_threads.append(t)
                t.start()

                logger.info("Connecting to %s" % (client.integration_name))

                # Have the client connect to the client's services
                if client.connect():
                    logger.info("%s client is connected" % (client.integration_name))

                    self.queue.put(client)
                else:
                    logger.error("Unable to connect to the client %s" % (client.integration_name))

            # Main thread of the bot
            self.process()
        else:
            logger.critical("No valid client integrations found. Make sure the settings.ini file has an entry for clients_enabled and that the settings for the client are configured")

    def process(self):
        try:
            while threading.activeCount() > 0:
                # logger.debug("Main Loop - active threads: %s, should_shutdown: %s, should_restart: %s" % (threading.activeCount(), self.should_shutdown, self.should_restart))
                # If exit mode is true the bot will be expecting porcessing threads to die off
                # During every loop the bot will be checking to see if background threads have
                # switch to inactive status. If so, the thread will be removed from the active
                # processing thread list
                if self.should_shutdown:
                    #  Updating the list with only the threads that are still active
                    self.processing_threads = [x for x in self.processing_threads if x.is_alive()]

                    # Integrations like xmpp rely on sleekxmpp behind the scenes for processing.
                    # Sleekxmpp spawns its own background threads for scheduling purposes that
                    # take a little while longer to stop then the items in self.processing_threads
                    # so this next conditional will wait not only until HavocBot threads are burned
                    # down but also until the total active threads are down to just the main thread.
                    # This probably should be redone to be cleaner. It is a TODO
                    if len(self.processing_threads) == 0:
                        if threading.activeCount() == 1:
                            logger.debug("Only the main thread is active. All background threads have exited")
                            self.should_shutdown = False
                        else:
                            logger.debug("Waiting on non HavocBot background thread to exit")
                    else:
                        logger.debug("Waiting on HavocBot background thread to exit")
                else:
                    # Reconfigure and restart the bot if coming from a restart event
                    if self.should_restart:
                        self.should_restart = False
                        self.configure()
                        self.start()
                time.sleep(1)
                pass
        except (KeyboardInterrupt, SystemExit) as e:
            logger.info("Interrupt received - %s" % (e))
            # Cleanup before finally exiting
            self.should_shutdown = True
            self.exit()

    def register_triggers(self, trigger_tuple_list):
        if trigger_tuple_list:
            # Make a shallow copy of the currently known triggers
            working_copy_triggers = copy.copy(self.triggers)
            working_copy_triggers += trigger_tuple_list

            triggers_length = len(trigger_tuple_list)
            triggers_phrase = "trigger" if len(trigger_tuple_list) == 1 else "triggers"
            existing_triggers_length = len(self.triggers)
            existing_triggers_phrase = "trigger" if len(self.triggers) == 1 else "triggers"

            logger.debug("Loading %s new %s. %s %s previously loaded" % (triggers_length, triggers_phrase, existing_triggers_length, existing_triggers_phrase))
            self.triggers = working_copy_triggers

    def unregister_triggers(self, trigger_tuple_list):
        if trigger_tuple_list:
            # Make a shallow copy of the currently known triggers
            working_copy_triggers = copy.copy(self.triggers)
            working_copy_triggers = [x for x in working_copy_triggers not in trigger_tuple_list]

            triggers_length = len(trigger_tuple_list)
            triggers_phrase = "trigger" if len(trigger_tuple_list) == 1 else "triggers"
            existing_triggers_length = len(self.triggers)
            existing_triggers_phrase = "trigger" if len(self.triggers) == 1 else "triggers"

            logger.debug("Removing %s existing %s. %s %s previously loaded" % (triggers_length, triggers_phrase, existing_triggers_length, existing_triggers_phrase))
            self.triggers = working_copy_triggers

    def reload_plugins(self):
        self.plugins_core = pluginmanager.load_plugins_core(self)
        self.plugins_custom = pluginmanager.load_plugins_custom(self)

    def exit(self):
        self.shutdown()
        sys.exit(0)

    def restart(self):
        self.should_restart = True
        self.shutdown()

    def shutdown(self):
        self.disconnect()

        # Need to kill threads
        for thread in self.processing_threads:
            if thread.is_active:
                thread.queue.task_done()
                thread.is_active = False

        self.clients = []
        self.plugins_core = []
        self.plugins_custom = []
        self.triggers = []
        self.is_configured = False

    def disconnect(self):
        self.should_shutdown = True
        for client in self.clients:
            logger.info("Disconnecting client %s" % (client.integration_name))
            client.disconnect()

    def signa_handler(self, signal, frame):
        logger.info("Received an interrupt. Starting shutdown")
        self.exit()

    def show_threads(self):
        for thread in self.processing_threads:
            logger.debug("HavocBot.show_threads() - %s - thread is %s. is_active set to %s, is_alive set to %s" % (len(self.processing_threads), thread, thread.is_active, thread.is_alive()))


class ClientThread(threading.Thread):
    def __init__(self, havocbot):
        threading.Thread.__init__(self)
        self.havocbot = havocbot
        self.queue = havocbot.queue
        self.is_active = True

    def run(self):
        while not self.havocbot.should_shutdown:
            try:
                client = self.queue.get()
                # Trigger the client integration's process() method
                client.process()
            except Exception as e:
                logger.error(e)
            finally:
                if self.is_active:
                    self.queue.task_done()
                    self.is_active = False



from abc import ABCMeta, abstractmethod, abstractproperty


class Client(object):
    __metaclass__ = ABCMeta

    @abstractproperty
    def integration_name(self):
        """ A name for the client integration.

        Args:
            self (Client): the HavocBotPlugin subclass
        Returns:
            A string that provides the name of the integration
        """
        pass

    @abstractmethod
    def configure(self, settings):
        """ Reads a settings bundle and configures the integration prior to startup.

        The settings bundle comes from reading the settings.ini file and each
        section is converted into a list of tuples

        Args:
            self (Client): the HavocBotPlugin subclass
            settings (list): a list of key-value tuples

            example:
            [('username', 'havocbot'), ('password', 'hunter9')]
        Returns:
            True/False depending on whether the client was able to get configured
        """
        pass

    @abstractmethod
    def connect(self):
        """ Connect to the integration client.

        Args:
            self (Client): the HavocBotPlugin subclass
        Returns:
            True/False depending on whether the client was able to connect
        """
        pass

    @abstractmethod
    def disconnect(self):
        """ Disconnect from the integration client.

        Args:
            self (Client): the HavocBotPlugin subclass
        """
        pass

    @abstractmethod
    def process(self):
        """ Processeses incoming and outgoing events while connected to
        the client integration.

        This is the main loop of the chat client that will process all events
        and will generally call the handle_message method to act on an event

        Args:
            self (Client): the HavocBotPlugin subclass
        """
        pass

    @abstractmethod
    def handle_message(self, **kwargs):
        """ Receives an event/message and processeses it as needed

        This is the main processing center of the chat client. It will receive
        nearly all the events direct from the method process() and will determine
        if the event is to be handled by a plugin by iterating through the known
        triggers and checking if there is a regex match against the trigger. If
        a match is found, the function for the trigger is called with the following
        arguments: self (Client subclass), message_object (Message subclass), and
        an optional capture_groups kwarg that includes the match.groups()

        Args:
            self (Client): the HavocBotPlugin subclass
            **kwargs: message_object (Message subclass)
        """
        pass

    @abstractmethod
    def send_message(self, **kwargs):
        """ Sends a single line message

        Args:
            self (Client): the HavocBotPlugin subclass
            **kwargs: channel (str), message (Message subclass instance)
        Raises:
            AttributeError: Client is likely not connected
            Exception: Unknown error when sending the message
        """
        pass

    @abstractmethod
    def send_messages_from_list(self, **kwargs):
        """ Sends a multi line message

        Takes in a list of Message subclass instances and forms them into
        a multi line message dependant on the rules of the client integration.
        This is often as simple as "\n".join(message)

        Args:
            self (Client): the HavocBotPlugin subclass
            **kwargs: channel (str), message (list of Message subclass instance)
        Raises:
            AttributeError: Client is likely not connected
            Exception: Unknown error when sending the message
        """
        pass



import logging

logger = logging.getLogger(__name__)


class Message(object):
    def __init__(self, text, user, channel, type_):
        self.text = text
        self.user = user
        self.channel = channel
        self.type_ = type_

    def __str__(self):
        return "Message(Text: '%s', User: '%s', Channel: '%s', Type: '%s')" % (self.text, self.user, self.channel, self.type_)



import logging

logger = logging.getLogger(__name__)


class User(object):
    def __init__(self, name, username):
        self.name = name
        self.username = username

    def __str__(self):
        return "User(Name: '%s', Username: '%s')" % (self.name, self.username)



from havocbot.client import Client
from havocbot.message import Message
from havocbot.user import User
import logging
import re
import sleekxmpp

logger = logging.getLogger(__name__)


class XMPP(Client):

    @property
    def integration_name(self):
        return "xmpp"

    def __init__(self, havocbot):
        # Capture a reference to havocbot
        self.havocbot = havocbot

        self.token = None
        self.username = None
        self.room_name = None
        self.nickname = None
        self.server = None
        self.exact_match_one_word_triggers = True

    # Takes in a list of kv tuples in the format [('key', 'value'),...]
    def configure(self, settings):
        for item in settings:
            # Switch on the key
            if item[0] == 'jabber_id':
                self.username = item[1]
            elif item[0] == 'password':
                self.password = item[1]
            elif item[0] == 'room_name':
                self.room_name = item[1]
            elif item[0] == 'nickname':
                self.nickname = item[1]
            elif item[0] == 'server':
                self.server = item[1]

        # Return true if this integrations has the information required to connect
        if self.username is not None and self.password is not None and self.room_name is not None and self.nickname is not None and self.server is not None:
            return True
        else:
            logger.error("XMPP configuration is not valid. Check your settings and try again")
            return False

    def connect(self):
        if self.username is None:
            logger.error("A XMPP username must be configured")
        if self.password is None:
            logger.error("A XMPP password must be configured")

        self.client = MUCBot(self, self.havocbot, self.username, self.password, self.room_name + "@" + self.server, self.nickname)

        self.client.register_plugin('xep_0030')  # Service Discovery
        self.client.register_plugin('xep_0045')  # Multi-User Chat
        self.client.register_plugin('xep_0199', {'keepalive': True, 'interval': 60})  # XMPP Ping set for a keepalive ping every 60 seconds

        if self.client.connect():
            logger.info("I am.. %s! (%s)" % (self.nickname, self.username))
            return True
        else:
            return False

    def disconnect(self):
        if self.client is not None:
            self.client.disconnect()

    def process(self):
        self.client.process(block=True)

    def handle_message(self, **kwargs):
        if kwargs is not None:
            if 'message_object' in kwargs and kwargs.get('message_object') is not None:
                message_object = kwargs.get('message_object')

            if message_object.type_ == 'message':
                for (trigger, triggered_function) in self.havocbot.triggers:
                    # Add exact regex match if user defined
                    if len(trigger.split()) == 1 and self.exact_match_one_word_triggers is True:
                        if not trigger.startswith('^') and not trigger.endswith('$'):
                            logger.debug("Converting trigger to a line exact match requirement")
                            trigger = "^" + trigger + "$"

                    # Use trigger as regex pattern and then search the message for a match
                    regex = re.compile(trigger)

                    match = regex.search(message_object.text)
                    if match is not None:
                        logger.info("Matched message against trigger '%s'" % (trigger))

                        # Pass the message to the function associated with the trigger
                        try:
                            triggered_function(self, message_object, capture_groups=match.groups())
                        except Exception as e:
                            logger.error(e)

                        break
                    else:
                        logger.debug("No match. Skipping '%s'" % (trigger))
                        pass
            else:
                logger.debug("Ignoring non message event of type '%s'" % (message_object.type_))

    def send_message(self, **kwargs):
        if kwargs is not None:
            if 'message' in kwargs and kwargs.get('message') is not None:
                message = kwargs.get('message')
            if 'channel' in kwargs and kwargs.get('channel') is not None:
                channel = kwargs.get('channel')

            if channel and message:
                logger.info("Sending message '%s' to channel '%s'" % (message, channel))
                try:
                    self.client.send_message(mto=channel, mbody=message, mtype='groupchat')
                except AttributeError:
                    logger.error("Unable to send message. Are you connected?")
                except Exception as e:
                    logger.error("Unable to send message. %s" % (e))

    def send_messages_from_list(self, **kwargs):
        if kwargs is not None:
            if 'message' in kwargs and kwargs.get('message') is not None:
                message = kwargs.get('message')
            if 'channel' in kwargs and kwargs.get('channel') is not None:
                channel = kwargs.get('channel')

            if channel and message:
                joined_message = "\n".join(message)
                logger.info("Sending message list '%s' to channel '%s'" % (joined_message, channel))
                try:
                    self.client.send_message(mto=channel, mbody=joined_message, mtype='groupchat')
                except AttributeError:
                    logger.error("Unable to send message. Are you connected?")
                except Exception as e:
                    logger.error("Unable to send message. %s" % (e))

    def get_user_by_id(self, name):
        room_full = "%s@%s" % (self.room_name, self.server)
        jabber_id = self.client.plugin['xep_0045'].getJidProperty(room_full, name, 'jid')

        user = create_user_object_from_json(name, jabber_id)

        if user:
            return user
        else:
            return None


class MUCBot(sleekxmpp.ClientXMPP):
    def __init__(self, callback, havocbot, jid, password, room, nick):
        sleekxmpp.ClientXMPP.__init__(self, jid, password)

        self.havocbot = havocbot
        self.parent = callback
        self.room = room
        self.nick = nick

        self.add_event_handler("session_start", self.start)
        self.add_event_handler("groupchat_message", self.muc_message)

    def start(self, event):
        self.get_roster()
        self.send_presence()
        self.plugin['xep_0045'].joinMUC(self.room, self.nick, wait=True)

    def muc_message(self, msg):
        if msg['mucnick'] != self.nick:
            message_object = Message(msg['body'], msg['mucnick'], self.room, "message")
            logger.info("Received - %s" % (message_object))

            try:
                self.parent.handle_message(message_object=message_object)
            except Exception as e:
                logger.error(e)


class XMPPUser(User):
    def __init__(self, real_name, jabber_id):
        self.name = real_name
        self.username = jabber_id

    def __str__(self):
        return "XMPPUser(Name: '%s', Username: '%s')" % (self.name, self.username)


# Returns a newly created user from a json source
def create_user_object_from_json(name, jabber_id):
    user = XMPPUser(name, jabber_id)
    logger.debug(user)

    return user



from abc import ABCMeta, abstractmethod, abstractproperty


class HavocBotPlugin(object):
    __metaclass__ = ABCMeta

    @abstractproperty
    def plugin_description(self):
        """ A brief description of the plugin.

        Read only property that provides an overall description of the plugin.
        This field will be read by the HelpPlugin to provide helpful trigger context

        Args:
            self (HavocBotPlugin): the HavocBotPlugin subclass
        Returns:
            A string that describes the purpose of the plugin
        """
        pass

    @abstractproperty
    def plugin_short_name(self):
        """ A short user friendly name for the plugin.

        Args:
            self (HavocBotPlugin): the HavocBotPlugin subclass
        Returns:
            A string that provides a short user friendly name of the plugin
        """
        pass

    @abstractproperty
    def plugin_usages(self):
        """ Lists usages of the plugin.

        Read only property that defines a grouping of trigger functionality for help
        purposes. This field will be read by the HelpPlugin to provide helpful trigger
        context

        Args:
            self (HavocBotPlugin): the HavocBotPlugin subclass
        Returns:
            A tuple of tuples where the child tuple contains exactly 3 items where:
            1 is a short string that states an available trigger
            2 is a short string that states an example usage of the trigger (can be None)
            3 is a short string that states the purpose of the trigger

            example:
            return (
                ("!help", None, "display basic help information"),
                ("!help plugins", None, "list the enabled plugins"),
                ("!help plugins <plugin>", None, "display detailed information about a specific plugin"),
            )
        """
        pass

    @abstractproperty
    def plugin_triggers(self):
        """ Triggers that will activate the plugin.

        Read only property that defines what the bot will listen for in chats and
        what the bot will call if the event is to be handled.

        Args:
            self (HavocBotPlugin): the HavocBotPlugin subclass
        Returns:
            A string that describes the plugin
        """
        pass

    @abstractmethod
    def init(self, havocbot):
        """ Sets up the basic requirements for a plugin.

        Sets up the triggers dictionary and help dictionary

        Args:
            self (HavocBotPlugin): The HavocBotPlugin subclass
            havocbot (HavocBot): A havocbot instance
        """
        pass

    @abstractmethod
    def shutdown(self):
        """ Performs any work when a shutdown event is triggered.

        This may be called as part of a restart event and should handle any cleanup
        to get the plugin into a restartable state

        Args:
            self (HavocBotPlugin): The HavocBotPlugin subclass
        """
        pass

    @abstractmethod
    def start(self, callback, message, **kwargs):
        """ Must deal with the default entry point for an activated plugin.

        Args:
            self (HavocBotPlugin): The HavocBotPlugin subclass
            callback (Client): A havocbot client instance
            message (Message): A havocbot message subclass instance
            **kwargs: any addional keyword arguments supplied by the client instance
        """
        pass



from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()\

setup(
    name='havocbot',
    version='0.1.1',
    description='An extensible chat bot that works across any number of chat programs',
    long_description=long_description,
    url='https://github.com/markperdue/havocbot',
    author='Mark Perdue',
    author_email='markaperdue@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Communications :: Chat',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.0',
    ],
    keywords=['chat', 'slack', 'jabber', 'xmpp', 'hipchat'],
    packages=find_packages('src', exclude=['havocbot.plugins', 'havocbot.plugins.*']),
    package_dir={'': "src"},
    zip_safe=False,
    install_requires=['requests>=2.6.0', 'sleekxmpp>=1.3.1', 'slackclient>=0.16', 'python-dateutil>=1.4'],
)



from havocbot.common import catch_exceptions
from havocbot.plugin import HavocBotPlugin
import imp
import logging
import os

logger = logging.getLogger(__name__)


class StatefulPlugin:
    def __init__(self, havocbot, name, path):
        self.data = None
        self.path = path
        self.name = name
        self.handler = None
        self.init(havocbot)

    # Load a havocbot plugin
    # The handler is set to the exported class instance of the plugin
    @catch_exceptions
    def init(self, havocbot):
        plugin = imp.load_source(self.name, self.path)
        self.handler = plugin.havocbot_handler

        # Call the init method in the plugin
        self.handler.init(havocbot)

    # Trigger the shutdown of the plugin
    @catch_exceptions
    def shutdown(self):
        self.handler.shutdown()

    @catch_exceptions
    def handle(self, message):
        return self.handler.handle_message(message)

    # Determines if the object at a path is a havocbot plugin
    @staticmethod
    def is_valid(path):
        if not path.endswith(".py"):
            return False

        f = open(path, "rb")
        data = f.read(16)
        f.close()
        if data.startswith(b"#!/havocbot"):
            return True

        return False


# Load a plugin by name
def load_plugin(havocbot, name, path):
    logger.log(0, "load_plugin called with name '%s' and path '%s'" % (name, path))

    if StatefulPlugin.is_valid(path):
        logger.debug("%s validated as a plugin" % (name))

        return StatefulPlugin(havocbot, name, path)
    else:
        return None


# Load all core plugins and attaches them to the bot
def load_plugins_core(havocbot):
    logger.log(0, "load_plugins_core triggered")

    # Trigger shutdown of any running plugins
    unload_plugins_of_type(havocbot, 'core')

    return load_plugins_of_type(havocbot, 'core')


# Load all custom plugins and attaches them to the bot
def load_plugins_custom(havocbot):
    # Trigger shutdown of any running plugins
    unload_plugins_of_type(havocbot, 'custom')

    return load_plugins_of_type(havocbot, 'custom')


# Load plugins of a type
def load_plugins_of_type(havocbot, plugin_type):
    plugins = []

    if plugin_type == "core":
        # Must load through pkg_resources since the bot may have been setup through pip and full filepaths for resources may not exist
        # http://peak.telecommunity.com/DevCenter/PkgResources#resource-extraction
        import pkg_resources
        core_package = 'havocbot.core'
        resources_list = pkg_resources.resource_listdir(core_package, '')

        for f in resources_list:

            # Remove file extension
            name, ext = os.path.splitext(f)

            if ext == '.py':
                # TODO - Optimize this. resource_filename is slow
                resource_filename = pkg_resources.resource_filename(core_package, f)

                plugin = load_plugin(havocbot, name, resource_filename)
                if plugin and isinstance(plugin.handler, HavocBotPlugin):
                    logger.info("%s core plugin loaded" % (name))
                    plugins.append(plugin)
    elif plugin_type == "custom":
        for listing in havocbot.plugin_dirs:
            folder = os.path.abspath(listing)
            if os.path.isdir(folder):
                for f in os.listdir(folder):
                    fpath = os.path.join(folder, f)

                    # Remove file extension
                    body, ext = os.path.splitext(f)

                    plugin = load_plugin(havocbot, body, fpath)
                    if plugin and isinstance(plugin.handler, HavocBotPlugin):
                        logger.info("%s custom plugin loaded" % (body))
                        plugins.append(plugin)
            else:
                logger.error("Plugin directory '%s' was listed in the settings file but does not exist" % (listing))

    return plugins


# Triggers the shutdown method in all loaded plugins
def unload_plugins_of_type(havocbot, plugin_type):
    if plugin_type == "core":
        for plugin in havocbot.plugins_core:
            plugin.handler.shutdown()
    elif plugin_type == "custom":
        for plugin in havocbot.plugins_custom:
            plugin.handler.shutdown()
    else:
        return
